/*
 * Contiguous Memory Allocator for DMA mapping framework
 * Copyright (c) 2010-2011 by Samsung Electronics.
 * Written by:
 *	Marek Szyprowski <m.szyprowski@samsung.com>
 *	Michal Nazarewicz <mina86@mina86.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License or (at your optional) any later version of the license.
 */

#define pr_fmt(fmt) "\033[31mFunction = %s, Line = %d, cma: " fmt "\033[m\n", __PRETTY_FUNCTION__, __LINE__

#ifdef CONFIG_CMA_DEBUG
#ifndef DEBUG
#  define DEBUG
#endif
#endif


#include <linux/memblock.h>
#include <linux/err.h>
#include <linux/mm.h>
#include <linux/mutex.h>
#include <linux/page-isolation.h>
#include <linux/slab.h>
#include <linux/swap.h>
#include <linux/mm_types.h>
#include <linux/cma_allocator.h>
#include <asm/page.h>
#include <linux/compaction.h>
#include <linux/migrate.h>
#include <linux/freezer.h>
#include <linux/bug.h>
#include <linux/kthread.h>

#ifndef CONFIG_COMPACTION
#error You should enable CONFIG_COMPACTION in kernel features to support the CMA feature!
#endif

struct cma {
	unsigned long	base_pfn;
	unsigned long	count;
	unsigned long	*bitmap;
	unsigned long   zero_bit_count;
	unsigned long  *pcache_low_threshold;			//point to cma_miuX_threshold[0]
	unsigned long  *pcache_high_threshold;		//point to cma_miuX_threshold[1]
	unsigned long  cache_increase_decrease_batch;
	unsigned long  last_free_high;
	unsigned long  lock_bits;

	wait_queue_head_t cma_swap_wait;//page swap worker wait queue
	struct task_struct *swap_worker;
	
	spinlock_t      free_list_spinlock;//spin lock to protect free page list
	unsigned long  free_page_count;
	struct list_head      free_list;
	struct zone    *zone;
} *mali_cma_ptr[2];

atomic_t g_mali_alloc_total,g_mali_free_total,g_fallback_total,g_miu0_fail,g_miu1_fail;
static unsigned long cma_reserve_addr[2] = {0,0};
static unsigned long cma_reserve_size[2] = {0,0};

static struct shrinker cma_shrinker;
extern unsigned long mali_alloc_strategy[2];
extern unsigned long cma_miu0_threshold[3];
extern unsigned long cma_miu1_threshold[3];
extern unsigned long miu0_free_in_cache;
extern unsigned long miu1_free_in_cache;

#define CMA_BUG_ON(condition, format...) ({						\
	int __ret_warn_on = !!(condition);				\
	if (unlikely(__ret_warn_on))					\
		__WARN_printf(format);					\
	if (unlikely(__ret_warn_on))					\	
	BUG();					\
})

extern unsigned long reclaim_clean_pages_from_list(struct zone *zone,
							struct list_head *page_list);
#if MP_MM_CMA_DEBUG==1	
static void check_freecount_coherent(struct cma *cma);
#endif


static int __init early_cma_miu0(char *p)
{
    if( p != NULL )
    {
        sscanf(p,"%lx,%lx",&cma_reserve_addr[0],&cma_reserve_size[0]);
    }
    else
    {
        printk("cma_miu0 not set\n");
    }
    return 0;
}
static int __init early_cma_miu1(char *p)
{
    if( p != NULL )
    {
        sscanf(p,"%lx,%lx",&cma_reserve_addr[1],&cma_reserve_size[1]);
    }
    else
    {
        printk("cma_miu0 not set\n");
    }
    return 0;
}

early_param("cma_miu0", early_cma_miu0);
early_param("cma_miu1", early_cma_miu1);


static struct cma_reserved {
	phys_addr_t start;
	unsigned long size;
} cma_reserved[2] __initdata;
//static unsigned cma_reserved_count __initdata;	// cma_reserved_count may not be 0, this means some cma is pre-reserved, cma_init_reserved_areas will clear this pre-reserved area


void get_mali_alloc_strategy(unsigned long *ump, unsigned long * mali)
{
	if(!ump || !mali){
		printk("invalid arg!!! \n");
	}
		
	*ump = mali_alloc_strategy[0];
	*mali = mali_alloc_strategy[1];	
}
EXPORT_SYMBOL(get_mali_alloc_strategy);


void set_mali_alloc_strategy(unsigned long ump, unsigned long mali)
{
	//printk("set mali_alloc_strategy from %d %d to %d %d \n",mali_alloc_strategy[0],mali_alloc_strategy[1],ump,mali);
	mali_alloc_strategy[0] = ump;
	mali_alloc_strategy[1] = mali;	
}
EXPORT_SYMBOL(set_mali_alloc_strategy);


void get_free_in_cache(unsigned long* miu0_free,unsigned long* miu1_free)
{
	if(mali_cma_ptr[0])
		*miu0_free = mali_cma_ptr[0]->free_page_count;
	else
		*miu0_free = 0;
	
	if(mali_cma_ptr[1])
		*miu1_free = mali_cma_ptr[1]->free_page_count;
	else
		*miu1_free = 0;
}


/**
 * cma_declare_contiguous() - reserve area for contiguous memory handling
 *			      for particular device
 * @dev:   Pointer to device structure.
 * @size:  Size of the reserved memory.
 * @base:  Start address of the reserved memory (optional, 0 for any).
 * @limit: End address of the reserved memory (optional, 0 for any).
 *
 * This function reserves memory for specified device. It should be
 * called by board specific code when early allocator (memblock or bootmem)
 * is still activate.
 */
static int __init cma_declare_contiguous(struct cma_reserved *r, phys_addr_t base, phys_addr_t size)
{
	phys_addr_t alignment;

	pr_debug("%s(size %lx, base %08lx)\n", __func__,
		 (unsigned long)size, (unsigned long)base);
	


	if (!size)
		return -EINVAL;

	/* Sanitise input arguments */
	alignment = PAGE_SIZE << max(MAX_ORDER - 1, pageblock_order);	// 4MB alignment
	size = round_down(base+size, alignment);	// 							  0x200000 ->   0x400000
	base = round_up(base, alignment);	// to next 4MB address, ex: 0x20200000 -> 0x20400000
	if(size <=base)
		return -EINVAL;
	size -= base;
	

	/* Reserve memory */

	printk("\033[35mFunction = %s, Line = %d, base is 0x%X\033[m\n", __PRETTY_FUNCTION__, __LINE__, (unsigned int)base);
	if (memblock_is_region_reserved(base, size) || memblock_reserve(base, size) < 0)
	{
		printk("\033[35mFunction = %s, Line = %d\033[m\n", __PRETTY_FUNCTION__, __LINE__);
		base = -EBUSY;
		goto err;
	}

	/*
	 * Each reserved area must be initialised later, when more kernel
	 * subsystems (like slab allocator) are available.
	 */
	r->start = base;
	r->size = size;
	pr_info("CMA: reserved %ld MiB at %08lx\n", (unsigned long)size / SZ_1M,
		(unsigned long)base);
	
	return 0;
err:
	pr_err("CMA: failed to reserve %ld MiB\n", (unsigned long)size / SZ_1M);
	return base;
}


/**
 * cma_contiguous_reserve() - reserve area for contiguous memory handling
 *
 * This function reserves memory from early allocator. It should be
 * called by arch specific code once the early allocator (memblock or bootmem)
 * has been activated and all other subsystems have already allocated/reserved
 * memory.
 */
void __init cma_contiguous_reserve(void)
{

	pr_debug("%s\n", __func__);
	
	cma_reserved[0].size = 0;
	cma_reserved[1].size = 0;

	if (cma_reserve_addr[0] && cma_reserve_size[0]) {
		pr_info("%s: reserving [addr 0x%lx, size %ld \n", __func__,
			 cma_reserve_addr[0], cma_reserve_size[0]);

		cma_declare_contiguous(&cma_reserved[0], cma_reserve_addr[0], cma_reserve_size[0]);
	}
	if (cma_reserve_addr[1] && cma_reserve_size[1]) {
		pr_info("%s: reserving [addr 0x%lx, size %ld \n", __func__,
			 cma_reserve_addr[1], cma_reserve_size[1]);

		cma_declare_contiguous(&cma_reserved[1], cma_reserve_addr[1], cma_reserve_size[1]);
	}

};



static DEFINE_MUTEX(cma_mutex);

static __init int cma_activate_area(unsigned long base_pfn, unsigned long count)
{
	unsigned long pfn = base_pfn;
	unsigned i = count >> pageblock_order;	// one pageblock having 1024 pages(4MB), having how much pageblock
	struct zone *zone;

	WARN_ON_ONCE(!pfn_valid(pfn));
	zone = page_zone(pfn_to_page(pfn));
	printk("\033[35mFunction = %s, Line = %d, pfn: 0x%X is in zone %s\033[m\n", __PRETTY_FUNCTION__, __LINE__, (unsigned int)pfn, zone->name);

	do {
		unsigned j;
		base_pfn = pfn;

		// check all pages in this pageblock are vaild and in the same zone 
		for (j = pageblock_nr_pages; j; --j, pfn++) {
			WARN_ON_ONCE(!pfn_valid(pfn));
			if (page_zone(pfn_to_page(pfn)) != zone)
				return -EINVAL;
		}
		//printk("\033[35mFunction = %s, Line = %d, pfn is 0x%X\033[m\n", __PRETTY_FUNCTION__, __LINE__, (unsigned int)pfn);
		//printk("\033[35mFunction = %s, Line = %d, base_pfn is 0x%X\033[m\n", __PRETTY_FUNCTION__, __LINE__, (unsigned int)base_pfn);
		init_cma_reserved_pageblock(pfn_to_page(base_pfn));
	} while (--i);
	return 0;
}

static __init struct cma *cma_create_area(unsigned long base_pfn,
				     unsigned long count)
{
	int bitmap_size = BITS_TO_LONGS(count) * sizeof(long);	// count is page_num, BITS_TO_LONGS(count) coverts count to ~ long size. bitmap_size is ~ byte the bitmap needs
	struct cma *cma;
	int ret = -ENOMEM;

	pr_info("%s(base %08lx, count %lx, bitmap_size %d(bytes))\n", __func__, base_pfn, count, bitmap_size);

	cma = kmalloc(sizeof *cma, GFP_KERNEL);
	if (!cma)
		return ERR_PTR(-ENOMEM);

	cma->base_pfn = base_pfn;
	cma->count = count;
	cma->bitmap = kzalloc(bitmap_size, GFP_KERNEL);

	if (!cma->bitmap)
		goto no_mem;

	ret = cma_activate_area(base_pfn, count);
	if (ret)
		goto error;

	pr_debug("%s: returned %p\n", __func__, (void *)cma);
	return cma;

error:
	kfree(cma->bitmap);
no_mem:
	kfree(cma);
	return ERR_PTR(ret);
}

static int mali_cma_pool_shrink(struct shrinker *shrinker, struct shrink_control *sc)
{
    struct cma *cma;
	int freed = 0;
	int target_miu;
	int total = 0;
	struct page *page;

	if (allocflags_to_migratetype(sc->gfp_mask) != MIGRATE_MOVABLE)
		return -1;

	if(mali_alloc_strategy[0])
	{
	   target_miu = 1;
	}
	else
	{
	   target_miu = 0;
	}

	if((cma=mali_cma_ptr[target_miu])==NULL)
		cma = mali_cma_ptr[target_miu^1];

	if(!cma)
		return 0;

	if(!mutex_trylock(&cma_mutex))
		goto OUT_NO_LOCK;
		
TRY_AGAIN:


	while(1)
	{
	    unsigned long pfn;

		spin_lock(&cma->free_list_spinlock);
		if(list_empty(&cma->free_list))
		{
		    spin_unlock(&cma->free_list_spinlock);
		    break;
		}
		page = list_entry(cma->free_list.next,
							  struct page, lru);
		list_del(&page->lru);
		cma->free_page_count--;
		
		spin_unlock(&cma->free_list_spinlock);
		pfn = page_to_pfn(page);
		
		CMA_BUG_ON(pfn<cma->base_pfn || pfn>=cma->base_pfn+cma->count,"pfn=%lx,cma[%lx,%lx] \n",pfn,cma->base_pfn,cma->count);
		
		if(!test_and_clear_bit(pfn - cma->base_pfn, cma->bitmap))
		{
			BUG();
		}
		cma->zero_bit_count++;
		CMA_BUG_ON(page_count(page) != 1,"page count=%d \n",page_count(page));
		__free_page(page);
		freed++;
		if(freed >= sc->nr_to_scan)
			break;
	}
#if MP_MM_CMA_DEBUG==1	
	check_freecount_coherent(cma);
#endif

    if(freed < sc->nr_to_scan)
    {
       if(cma == mali_cma_ptr[target_miu])
       {
          if((cma=mali_cma_ptr[target_miu^1])!=NULL)
		  	goto TRY_AGAIN;
       }
    }
	mutex_unlock(&cma_mutex);

OUT_NO_LOCK:
    total +=(mali_cma_ptr[0]?mali_cma_ptr[0]->free_page_count:0);
	total +=(mali_cma_ptr[1]?mali_cma_ptr[1]->free_page_count:0);
	
	return total;

}



/**
 * dma_alloc_from_contiguous() - allocate pages from contiguous area
 * @dev:   Pointer to device for which the allocation is performed.
 * @count: Requested number of pages.
 * @align: Requested alignment of pages (in PAGE_SIZE order).	// request 2^align pages
 *
 * This function allocates memory buffer for specified device. It uses
 * device specific contiguous memory area if available or the default
 * global one. Requires architecture specific get_dev_cma_area() helper
 * function.
 */


struct page *cma_alloc_from_contiguous(int count,
				       unsigned int align)
{
	unsigned long mask, pfn, pageno, start = 0;
	struct cma *cma;
	struct page *page = NULL;
	int ret;

	if(mali_alloc_strategy[0] && mali_cma_ptr[0])
		cma = mali_cma_ptr[0];
	else if(mali_alloc_strategy[1] && mali_cma_ptr[1])
		cma = mali_cma_ptr[1];
	else if(mali_cma_ptr[0])
		cma = mali_cma_ptr[0];
	else 
		cma = mali_cma_ptr[1];
	
	if (!cma || !cma->count)
		return NULL;


	pr_debug("%s(cma %p, count %d, align %d)\n", __func__, (void *)cma, count, align);

	if (!count)
		return NULL;

	mask = (1 << align) - 1;

	mutex_lock(&cma_mutex);

	for (;;) {
		// bitmap_find_next_zero_area will return an index which points a zero-bit in bitmap, whose following count bits are all zero-bit
		pageno = bitmap_find_next_zero_area(cma->bitmap, cma->count,
						    start, count, mask);
		if (pageno >= cma->count)	// no such continuous area
			break;

		pfn = cma->base_pfn + pageno;
		ret = alloc_contig_range(pfn, pfn + count, MIGRATE_CMA);
		if (ret == 0) {
			bitmap_set(cma->bitmap, pageno, count);
			page = pfn_to_page(pfn);
			break;
		} else if (ret != -EBUSY) {
			break;
		}
		pr_debug("%s(): memory range at %p is busy, retrying\n",
			 __func__, pfn_to_page(pfn));
		/* try again with a bit different memory target */
		start = pageno + mask + 1;
	}

	mutex_unlock(&cma_mutex);
	return page;
}

/**
 * dma_release_from_contiguous() - release allocated pages
 * @dev:   Pointer to device for which the pages were allocated.
 * @pages: Allocated pages.	// start pfn
 * @count: Number of allocated pages.
 *
 * This function releases memory allocated by dma_alloc_from_contiguous().
 * It returns false when provided pages do not belong to contiguous area and
 * true otherwise.
 */
static struct cma *get_cma(struct page *page)
{
	unsigned long pfn = page_to_pfn(page);

    if(mali_cma_ptr[0] && pfn>= mali_cma_ptr[0]->base_pfn &&
		   pfn< mali_cma_ptr[0]->base_pfn+mali_cma_ptr[0]->count)
    {
		   return mali_cma_ptr[0];
    }
	else if(mali_cma_ptr[1] && pfn>= mali_cma_ptr[1]->base_pfn &&
		   pfn< mali_cma_ptr[1]->base_pfn+mali_cma_ptr[1]->count)
    {
		   return mali_cma_ptr[1];
    }
	//printk(KERN_ERR"############### mali_cma_ptr %p,%p ; pfn=%x \n",mali_cma_ptr[0],mali_cma_ptr[1],(unsigned int)pfn);
	//if(mali_cma_ptr[0])
		//printk(KERN_ERR"############### [0]%p : base=%x,count=%lu \n",mali_cma_ptr[0],(unsigned int)mali_cma_ptr[0]->base_pfn,mali_cma_ptr[0]->count);
	//if(mali_cma_ptr[1])
		//printk(KERN_ERR"###############[1]%p : base=%x,count=%lu \n",mali_cma_ptr[1],(unsigned int)mali_cma_ptr[1]->base_pfn,mali_cma_ptr[1]->count);
		
	return NULL;

}
bool cma_release_from_contiguous(struct page *pages,
				 int count)
{
 	struct cma *cma = get_cma(pages);
	unsigned long pfn;

	printk("\033[31mFunction = %s, Line = %d\033[m\n", __PRETTY_FUNCTION__, __LINE__); // joe.liu

	if (!cma || !pages)
		return false;

	pr_debug("%s(page %p)\n", __func__, (void *)pages);

	pfn = page_to_pfn(pages);

	if (pfn < cma->base_pfn || pfn >= cma->base_pfn + cma->count)
		return false;

	VM_BUG_ON(pfn + count > cma->base_pfn + cma->count);

	mutex_lock(&cma_mutex);
	bitmap_clear(cma->bitmap, pfn - cma->base_pfn, count);
	free_contig_range(pfn, count);
	mutex_unlock(&cma_mutex);

	return true;
}

//should be called under cma_mutex locked
#if 1//MP_MM_CMA_DEBUG==1	
static unsigned long  get_free_bit_count(struct cma *cma, unsigned long start, unsigned long len)
{
    unsigned long count = 0;
	unsigned long offset_zero;
	unsigned long end = start+len;


    if(len==0 || start >= cma->count)
		return 0;
	end = start+len;
	end = min(end, cma->count);
	
	offset_zero = start;
	
    while(offset_zero < cma->count)
    {
        unsigned offset_bit;
	    offset_zero = find_next_zero_bit(cma->bitmap, end,
						 offset_zero);
	    if(offset_zero>= end)
	  	   break;
	    offset_bit = find_next_bit(cma->bitmap, end, offset_zero+1);
		count += offset_bit-offset_zero;
		offset_zero = offset_bit+1;
     }

   return count;
}
#endif

static int cma_cache_pages(unsigned long start, unsigned long end,
	        struct cma *cma, int target_count)
{
  int page_count;
  unsigned long last_free_high;
  struct page *page;
  
  struct list_head freelist;
  INIT_LIST_HEAD(&freelist);
  lru_add_drain_all();
  drain_all_pages();
  page_count = isolate_descrete_freepages(cma->zone, start, end, target_count, &freelist,
  	   &last_free_high);
  if(page_count)
  {
        int page_count2 = 0;
        list_for_each_entry(page, &freelist, lru) 
	    {
	   	  unsigned long pfn_tmp;
	   	  pfn_tmp = page_to_pfn(page);
		  CMA_BUG_ON(pfn_tmp<cma->base_pfn || pfn_tmp>=cma->base_pfn+cma->count,"pfn=%lx,cma[%lx,%lx] \n",pfn_tmp,cma->base_pfn,cma->count);
		  if(test_and_set_bit(pfn_tmp-cma->base_pfn, cma->bitmap))
		  {
		      BUG();//this bit should be zero before
		  }
		  cma->zero_bit_count--;
		  page_count2++;
        }
		CMA_BUG_ON(page_count != page_count2,"page_count=%d,page_count2=%d\n",page_count,page_count2);
		
		spin_lock(&cma->free_list_spinlock);
	    list_splice(&freelist, &cma->free_list);
		cma->free_page_count += page_count;
		spin_unlock(&cma->free_list_spinlock);
	
  }
  cma->last_free_high = last_free_high;
  return page_count;
}

static unsigned long pfn_max_align_down(unsigned long pfn)
{
	return pfn & ~(max_t(unsigned long, MAX_ORDER_NR_PAGES,
			     pageblock_nr_pages) - 1);
}

static unsigned long pfn_max_align_up(unsigned long pfn)
{
	return ALIGN(pfn, max_t(unsigned long, MAX_ORDER_NR_PAGES,
				pageblock_nr_pages));
}
int cma_increase_pool_freelist_threshold(struct cma *cma)
{
	unsigned long pfn, last_pfn, end, nr_reclaimed;
	int ret;
	int total_page_count = 0;
	int tries = 0;

	unsigned long isolated_beg=(unsigned long)-1, isolated_end=(unsigned long)-1;

	unsigned long last_search_beg;
	unsigned long page_count;
	
	struct compact_control cc = {
		.nr_freepages = 0,
		.nr_migratepages = 0,
		.order = 0,
		.migratetype = MIGRATE_CMA,
		.zone = cma->zone,
		.sync = true,
	};
	INIT_LIST_HEAD(&cc.freepages);
	INIT_LIST_HEAD(&cc.migratepages);

	mutex_lock(&cma_mutex);
	
	migrate_prep();
	pfn = cma->base_pfn;
	end = cma->base_pfn+cma->count;
	page_count = 0;
	last_pfn = (unsigned long)-1;

	while ( pfn < end|| !list_empty(&cc.migratepages)) {
		if (fatal_signal_pending(current)) {
			ret = -EINTR;
			break;
		}
		
		if (list_empty(&cc.migratepages)) {
			cc.nr_migratepages = 0;
							
			pfn = cma->base_pfn+
				 find_next_zero_bit(cma->bitmap, cma->count, pfn-cma->base_pfn);

next_block:
			if(pfn >= end)
				break;
			
			last_pfn = pfn;

			//no block isolated, isolated one
			if(isolated_beg == (unsigned long)-1){
				isolated_beg = pfn_max_align_down(pfn);
				isolated_end = pfn_max_align_up(pfn+1);
				ret = start_isolate_page_range(isolated_beg,
				                        isolated_end, MIGRATE_CMA);
#if 0			//code for debug usage	
				if(cma->free_page_count == 0 && get_free_bit_count(cma,isolated_beg-cma->base_pfn,1024) > 512)
					g_flag = 1;
				else
					g_flag = 0;
#endif				
				if(ret)
				{ 
				    //isolate the block failed
					printk(KERN_ERR" isolate failed [%x,%x] \n",isolated_beg,isolated_end);
				    isolated_beg=(unsigned long)-1;
					pfn = isolated_end + 1;
					goto next_block;
				}
			}

			//next COMPACT_CLUSTER_MAX pages till isolated_end
			pfn = isolate_migratepages_range(cc.zone, &cc,
							 pfn, isolated_end);
			if (!pfn) {
				ret = -EINTR;
				break;
			}
#if 0			
			if(g_flag == 1)
				printk("[%x,%x],%x ~ %x, cc.nr_migratepages=%d \n",isolated_beg, isolated_end,last_pfn,pfn,cc.nr_migratepages);
#endif
			tries = 0;
		} else if (++tries == 5) {
		     putback_lru_pages(&cc.migratepages);
			 CMA_BUG_ON(isolated_beg == (unsigned long)-1, "");

			 undo_isolate_page_range(isolated_beg, isolated_end, MIGRATE_CMA);
#if 0			 
			 printk("migrate fail!!! [%x,%x],%x ~ %x,migrate pages=%d,zero_bit_count=%d,total_free_count=%d,block_free=%d \n",
			 	isolated_beg, isolated_end,last_pfn,pfn,cc.nr_migratepages,cma->zero_bit_count,
			 	cma->free_page_count,get_free_bit_count(cma,isolated_beg-cma->base_pfn,1024));
#endif
			 isolated_beg = -1;
			 continue;
		}	
	
		if(!list_empty(&cc.migratepages)){
			nr_reclaimed = reclaim_clean_pages_from_list(cc.zone,
								&cc.migratepages);
			cc.nr_migratepages -= nr_reclaimed;
	
			ret = migrate_pages(&cc.migratepages, alloc_migrate_target,
						(unsigned long)&cc, false, true);
		}
		
		if(pfn >= isolated_end){	//end of a block
			int ret;
			//even though we get pages from [last_pfn, pfn), actually we get from the whole block
			total_page_count += ret = cma_cache_pages(last_pfn, pfn, cma, cma->cache_increase_decrease_batch);
#if 0			
			if(g_flag == 1)			
				printk("#######start %x,end %x, [%x,%x] getpages=%d,total_free_bit=%d ,total_free_page_count = %d,zerobit=%d, block_free=%d \n",last_pfn, pfn,isolated_beg, isolated_end,
					total_page_count,get_free_bit_count(cma,0,cma->count),
					cma->free_page_count,
					cma->zero_bit_count,
					get_free_bit_count(cma,isolated_beg-cma->base_pfn,1024));
#endif
			if(total_page_count > cma->cache_increase_decrease_batch)
				break;

			if(list_empty(&cc.migratepages)){	//empty, just goto next block
				//un-isolate the block
				if(isolated_beg != (unsigned long)-1)
					undo_isolate_page_range(isolated_beg, isolated_end, MIGRATE_CMA);
				isolated_beg = -1;			//isolated == -1 means no block isolated
				goto next_block;
			}
		}

	}

	 //clear migrate list
	 if(!list_empty(&cc.migratepages))
	 {
	     putback_lru_pages(&cc.migratepages);
	 }
	 if(isolated_beg !=(unsigned long)-1)
	 	  undo_isolate_page_range(isolated_beg, isolated_end, MIGRATE_CMA);

UNLOCK_RET:	
	mutex_unlock(&cma_mutex);
	if(total_page_count && waitqueue_active(&cma->cma_swap_wait))
		  wake_up_interruptible(&cma->cma_swap_wait);
	return total_page_count;
}

#if MP_MM_CMA_DEBUG==1	
static void check_freecount_coherent(struct cma *cma)
{
	struct list_head *next;
	int tmp=0;
	struct page *page;
	unsigned long pfn;

	spin_lock(&cma->free_list_spinlock);
	next = cma->free_list.next;
	while(next != &cma->free_list)
	{
	   page = container_of(next,struct page,lru);
	   pfn = page_to_pfn(page);

	   CMA_BUG_ON(pfn<cma->base_pfn || pfn>=cma->base_pfn+cma->count,"pfn=%x,cma[%x,%x] \n",pfn,cma->base_pfn,cma->count);
	   CMA_BUG_ON(!test_bit(pfn-cma->base_pfn,cma->bitmap),"1pfn=%x,cma[%x,%x] \n",pfn,cma->base_pfn,cma->count);
	   tmp++;
	   next = next->next;
	}
	CMA_BUG_ON(cma->free_page_count != tmp,"free_page_count=%d,tmp=%d\n",cma->free_page_count,tmp);
	spin_unlock(&cma->free_list_spinlock);

}
#endif


static int cma_swap_worker_thread(void *p)
{
  struct cma *cma = (struct cma *)p;
  
  set_freezable();

  while(1)
  {
     DEFINE_WAIT(wait);

#if MP_MM_CMA_DEBUG==1	
     //BUG_ON(zone_page_state(cma->zone, NR_FREE_PAGES)<zone_page_state(cma->zone, NR_FREE_CMA_PAGES));
     check_freecount_coherent(cma);
#endif
	 try_to_freeze();
	 
	 mutex_lock(&cma_mutex);
	 if(cma->free_page_count>*cma->pcache_high_threshold)
	 {
	    unsigned long pfn;
		struct page *page;
		
        while(cma->free_page_count>*cma->pcache_high_threshold-cma->cache_increase_decrease_batch)
        {
             spin_lock(&cma->free_list_spinlock);
             CMA_BUG_ON(list_empty(&cma->free_list),"");
		     page = list_entry(cma->free_list.next,
							   struct page, lru);
		     list_del(&page->lru);
			 cma->free_page_count--;
	
			 spin_unlock(&cma->free_list_spinlock);

			 pfn = page_to_pfn(page);
			 
			 CMA_BUG_ON(pfn<cma->base_pfn || pfn>=cma->base_pfn+cma->count,"pfn=%lx,cma[%lx,%lx] \n",pfn,cma->base_pfn,cma->count);
			 
			 if(!test_and_clear_bit(pfn - cma->base_pfn, cma->bitmap))
			 {
				BUG();
			 }
			 cma->zero_bit_count++;
			 CMA_BUG_ON(page_count(page) != 1,"page_count=%d\n",page_count(page));
		     __free_page(page);
 
        }
		
     }
	 mutex_unlock(&cma_mutex);

	 if(cma->free_page_count<*cma->pcache_low_threshold)
	 {
	     while(cma->free_page_count<*cma->pcache_low_threshold+cma->cache_increase_decrease_batch)
         {
	   	     if(!cma_increase_pool_freelist_threshold(cma))
			 	break;
         }
		 #if MP_MM_CMA_DEBUG==1	
         check_freecount_coherent(cma);
         #endif
	 }
     prepare_to_wait(&cma->cma_swap_wait, &wait, TASK_INTERRUPTIBLE);
	 schedule_timeout(HZ/10);
	 finish_wait(&cma->cma_swap_wait, &wait);
  }
  return 0;
}

atomic_t g_miu1_total,g_miu1_time,g_miu1_total_slow,g_miu1_time_slow;
//struct page *mali_alloc_page(unsigned long mali_mask)
struct page *mali_alloc_page()
{
	struct cma *cma;
	struct page *page;
	unsigned long before,fallback_flag=0,flag;

	atomic_inc(&g_mali_alloc_total);


#if 1	
	if(!mali_cma_ptr[1])
	   cma=mali_cma_ptr[0];
	else if(!mali_cma_ptr[0])
	   cma=mali_cma_ptr[1];
	
	if(mali_alloc_strategy[0])
	   cma = mali_cma_ptr[1];
	else
	   cma = mali_cma_ptr[0];
#else
	if(mali_mask & __GFP_UMP){
		if(mali_alloc_strategy[0])
			cma = mali_cma_ptr[1];
		else
			cma = mali_cma_ptr[0];
	}
	else if(mali_mask & __GFP_MALI){
		if(mali_alloc_strategy[1])
			cma = mali_cma_ptr[1];
		else
			cma = mali_cma_ptr[0];
	}
#endif
	
	if(!cma){
		fallback_flag = 1;
		goto FALLBACK_KERNEL_PAGE;
	}
	
TRY_AGAIN:
	spin_lock(&cma->free_list_spinlock);
	if(!list_empty(&cma->free_list))
	{
		page = list_entry(cma->free_list.next,
							   struct page, lru);
	  	list_del(&page->lru);
		
#if MP_MM_CMA_DEBUG==1	
		BUG_ON(!test_bit(page_to_pfn(page)-cma->base_pfn,cma->bitmap));
#endif
		cma->free_page_count--;
		spin_unlock(&cma->free_list_spinlock);
		if(cma->free_page_count<*cma->pcache_low_threshold)
		  wake_up_interruptible(&cma->cma_swap_wait);

		clear_highpage(page);
		return page;
	}

	spin_unlock(&cma->free_list_spinlock);

	before = jiffies;
	{ 
	   signed long timeout = HZ/5;
	   DEFINE_WAIT(wait);
	   if(cma->zero_bit_count <=100){//should refine condition here?
	   	  //printk(KERN_ERR"zero_bit_count=%d, free bit=%d ,times=%d\n",
		  	//cma->zero_bit_count,get_free_bit_count(cma, 0, cma->count),tmp);
		  fallback_flag = 1; 	
		  goto FALLBACK_KERNEL_PAGE;
	   }
	   try_to_freeze();

	   if(waitqueue_active(&cma->cma_swap_wait))
          wake_up_interruptible(&cma->cma_swap_wait);
	   
#if 0	   
	   while(timeout>=0)
	   {
	       prepare_to_wait(&cma->cma_swap_wait, &wait, TASK_INTERRUPTIBLE);
	       timeout -= (schedule_timeout(HZ/100)+1);
	       finish_wait(&cma->cma_swap_wait, &wait);
		   page = cma_get_cached_clean_page(cma);
		   if(cma->free_page_count == 0)
		   	printk(KERN_ERR"timeout = %d,timeused=%d,times=%d,pid=%d \n",
		   		timeout,jiffies-before,tmp,current->pid);
		   if(page)
			   return page;
	   }
#else
		while(timeout>=0){
			timeout -= (wait_event_timeout(cma->cma_swap_wait,cma->free_page_count > 0,HZ/100) + 1);
			
	       if(!list_empty(&cma->free_list))
	   	      goto TRY_AGAIN;	
		   if(waitqueue_active(&cma->cma_swap_wait))
		      wake_up_interruptible(&cma->cma_swap_wait);

		}
#endif
	}
		
FALLBACK_KERNEL_PAGE:
	if(mali_alloc_strategy[0] == 0)
		atomic_inc(&g_miu0_fail);
	else
		atomic_inc(&g_miu1_fail);	
	atomic_inc(&g_fallback_total);

	page =alloc_page(GFP_HIGHUSER| __GFP_ZERO | __GFP_REPEAT | __GFP_NOWARN);
	//printk(KERN_ERR"-------------fallback strategy[%d],pfn=%x, total=%lu,fallback=%lu \n",
		                           //tmp,(unsigned int)page_to_pfn(page),times,atomic_read(&g_fallback_total));

	 return page;

}
EXPORT_SYMBOL(mali_alloc_page);

void mali_free_page(struct page *page)
{
	struct cma *cma;

	cma = get_cma(page);

	atomic_inc(&g_mali_free_total);
	if(!cma)
	{
		__free_page(page);
		return;
	}
		
#if MP_MM_CMA_DEBUG==1	
	CMA_BUG_ON(!test_bit(page_to_pfn(page)-cma->base_pfn,cma->bitmap),"");
#endif	
    CMA_BUG_ON(atomic_read(&page->_count)!= 1,"page_count=%d \n",page_count(page));

	spin_lock(&cma->free_list_spinlock);
	list_add(&page->lru,&cma->free_list);
	cma->free_page_count++;
	spin_unlock(&cma->free_list_spinlock);
	
	if(cma->free_page_count>*cma->pcache_high_threshold
			&& waitqueue_active(&cma->cma_swap_wait))
	{
		wake_up_interruptible(&cma->cma_swap_wait);
	}
}
EXPORT_SYMBOL(mali_free_page);

#if MP_MM_CMA_DEBUG==1	
int cma_page_check(struct page* page)
{
	struct cma *cma;
	unsigned long pfn;

	cma = get_cma(page);
	if(!cma)
		return 1;
	
	pfn = page_to_pfn(page);
	if(test_bit(pfn-cma->base_pfn, cma->bitmap)){
		printk(KERN_ERR"pfn=%x,cma[%x,%x] \n",pfn,cma->base_pfn,cma->count);
		BUG();
	}
}
#endif

static int __init cma_init_reserved_areas(void)
{
	struct cma *cma;
	mali_cma_ptr[0] = NULL;
	mali_cma_ptr[1] = NULL;
	printk("\033[35mFunction = %s, Line = %d\033[m\n", __PRETTY_FUNCTION__, __LINE__);

    if(cma_reserved[0].size)
    {
	   cma = cma_create_area(PFN_DOWN(cma_reserved[0].start),
				      cma_reserved[0].size >> PAGE_SHIFT);
	   if (!IS_ERR(cma))
	   {
		    cma->zone = page_zone(pfn_to_page(cma->base_pfn));
			cma->zero_bit_count = cma->count;
			init_waitqueue_head(&cma->cma_swap_wait);
			if(cma->count/10 < cma_miu0_threshold[0])
				cma_miu0_threshold[0] = cma->count/10;
			cma->pcache_low_threshold = &cma_miu0_threshold[0];	//max 1M
			if(cma->count/4 < cma_miu0_threshold[1])
				cma_miu0_threshold[1] = cma->count/4;
			cma->pcache_high_threshold = &cma_miu0_threshold[1];	//max 32M
			cma->cache_increase_decrease_batch = cma_miu0_threshold[2];
			cma->free_page_count = 0;
			INIT_LIST_HEAD(&cma->free_list);
			cma->last_free_high = cma->base_pfn+cma->count-1;
			spin_lock_init(&cma->free_list_spinlock);
			mali_cma_ptr[0] = cma;
			cma->swap_worker = kthread_run(cma_swap_worker_thread, cma, "cma_swapper_0");
	        CMA_BUG_ON(IS_ERR(cma->swap_worker),"");
	   }
    }
    if(cma_reserved[1].size)
    {
	   cma = cma_create_area(PFN_DOWN(cma_reserved[1].start),
				      cma_reserved[1].size >> PAGE_SHIFT);
	   if (!IS_ERR(cma))
	   {
		    cma->zone = page_zone(pfn_to_page(cma->base_pfn));
		    cma->zero_bit_count = cma->count;
		    init_waitqueue_head(&cma->cma_swap_wait);
		    if(cma->count/10 < cma_miu1_threshold[0])
				cma_miu1_threshold[0] = cma->count/10;
		    cma->pcache_low_threshold = &cma_miu1_threshold[0];//max 1M
		    if(cma->count/4 < cma_miu0_threshold[1])
				cma_miu0_threshold[1] = cma->count/4;
		    cma->pcache_high_threshold = &cma_miu1_threshold[1];//max 32M
		    cma->cache_increase_decrease_batch = cma_miu1_threshold[2];
		    cma->free_page_count = 0;
			INIT_LIST_HEAD(&cma->free_list);
		    cma->last_free_high = cma->base_pfn+cma->count-1;
		    spin_lock_init(&cma->free_list_spinlock);
			mali_cma_ptr[1] = cma;
			cma->swap_worker = kthread_run(cma_swap_worker_thread, cma, "cma_swapper_1");
	        CMA_BUG_ON(IS_ERR(cma->swap_worker),"");
	   }
    }
	if(mali_cma_ptr[0] || mali_cma_ptr[1])
	{
	   cma_shrinker.shrink = mali_cma_pool_shrink;
	   cma_shrinker.seeks = DEFAULT_SEEKS;
	   cma_shrinker.batch = 0;
	   register_shrinker(&cma_shrinker);
	}
	
	return 0;
}
core_initcall(cma_init_reserved_areas);



